/*
  SDWebServer - Example WebServer with SD Card backend for esp8266

  Copyright (c) 2015 Hristo Gochkov. All rights reserved.
  This file is part of the WebServer library for Arduino environment.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  Have a FAT Formatted SD Card connected to the SPI port of the ESP8266
  The web root is the SD Card root folder
 

  upload the contents of SdRoot to the root of the SDcard and access the editor by going to http://esp8266sd.local/edit

*/
/*
  ESP32 S2 example to present USB MSC device to mount SD card
  SD card will be mount like a pendrive and should be formatted as FAT16/32
  in addition esp32 is using long names, up to 255 characters
*/
#include <WiFi.h>
#include <WiFiClient.h>
#include <WebServer.h>
#include <ESPmDNS.h>
#include <SPI.h>
#include <SD.h>
#include "mscusb.h"

#define DBG_OUTPUT_PORT Serial
#define HSPI_MISO   36
#define HSPI_MOSI   35
#define HSPI_SCLK   34
#define HSPI_SS     33

const char* ssid = "**********";
const char* password = "**********";
const char* host = "esp32sd";

WebServer server(80);

static bool hasSD = false;
File uploadFile;
SPIClass * hspi = NULL;

MSCusb msc;
uint32_t sectors;            // number of blocks in partition
uint16_t sector_size = 512;  // physical block size is 512 on SD card

// >>>>>>>>>>>>>>>> FAT partitions structures thanks to Mike Dunston (@atanisoft)
enum PARTITION_TYPE_T : uint8_t
{
  PART_EMPTY = 0x00,
  PART_FAT_12 = 0x01,
  PART_FAT_16 = 0x04,
/* -- unsupported currently.
  FAT_16B = 0x06,
  NTFS = 0x07,
  FAT_32_CHS = 0x0B,
  FAT_32_LBA = 0x0C,
  FAT_16B_LBA = 0x0E,
  EXTENDED = 0x0F,
*/
};

enum PARTITION_STATUS : uint8_t
{
  PART_STATUS_UNUSED = 0x00,
  PART_STATUS_ACTIVE = 0x80,
  PART_STATUS_BOOTABLE = 0x80
};

typedef struct TU_ATTR_PACKED
{
  PARTITION_STATUS  status;               //  1 status of the disk:
                                          //    0x00 = inactive
                                          //    0x01-0x7f = invalid
                                          //    0x80 = bootable
  uint8_t  first_head;                    //  2 
  uint8_t  first_sector;                  //  3 this field is split between sector and cylinder:
                                          //    bits 0-5 (0x3F) are for sector
                                          //    bits 6,7 are cylinder bits 8,9
  uint8_t  first_cylinder;                //  4
  PARTITION_TYPE_T  partition_type;       //  6
  uint8_t  last_head;                     //  7
  uint8_t  last_sector;                   //  8 this field is split between sector and cylinder:
                                          //    bits 0-5 (0x3F) are for sector
                                          //    bits 6,7 are cylinder bits 8,9
  uint8_t  last_cylinder;                 //  9
  uint32_t first_lba;                     // 12
  uint32_t sector_count;                  // 16
} partition_def_t;

typedef struct TU_ATTR_PACKED
{
  uint8_t bootstrap[218];                 // 218 -- unused today.
  uint16_t disk_timestamp;                // 220
  uint8_t original_drive_id;              // 221
  uint8_t disk_seconds;                   // 222
  uint8_t disk_minutes;                   // 223
  uint8_t disk_hours;                     // 224
  uint8_t boostrap2[216];                 // 440 NOTE: this may spill over into next two fields.
  uint32_t disk_signature;                // 444
  uint16_t copy_protected;                // 446
  partition_def_t partitions[4];          // 510
  uint8_t boot_signature[2];              // 512 -- boot sector signature
} master_boot_record_t;

// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

void returnOK() {
  server.send(204, "text/plain", "");
}

void returnFail(String msg) {
  server.send(500, "text/plain", msg + "\r\n");
}

bool loadFromSdCard(String path) {
  String dataType = "text/plain";
  if (path.endsWith("/")) {
    path += "index.html";
  }

  if (path.endsWith(".src")) {
    path = path.substring(0, path.lastIndexOf("."));
  } else if (path.endsWith(".html")) {
    dataType = "text/html";
  } else if (path.endsWith(".css")) {
    dataType = "text/css";
  } else if (path.endsWith(".js")) {
    dataType = "application/javascript";
  } else if (path.endsWith(".png")) {
    dataType = "image/png";
  } else if (path.endsWith(".gif")) {
    dataType = "image/gif";
  } else if (path.endsWith(".jpg")) {
    dataType = "image/jpeg";
  } else if (path.endsWith(".ico")) {
    dataType = "image/x-icon";
  } else if (path.endsWith(".xml")) {
    dataType = "text/xml";
  } else if (path.endsWith(".pdf")) {
    dataType = "application/pdf";
  } else if (path.endsWith(".zip")) {
    dataType = "application/zip";
  } else if (path.endsWith(".tbz")) {
    return "application/x-gzip";
  }

  File dataFile = SD.open(path.c_str());
  if (dataFile.isDirectory()) {
    path += "/index.html";
    dataType = "text/html";
    dataFile = SD.open(path.c_str());
  }

  if (!dataFile) {
    return false;
  }

  if (server.hasArg("download")) {
    dataType = "application/octet-stream";
  }

  if (server.streamFile(dataFile, dataType) != dataFile.size()) {
    DBG_OUTPUT_PORT.println("Sent less data than expected!");
  }

  dataFile.close();
  return true;
}

void handleFileUpload() {
  if (server.uri() != "/edit") {
    return;
  }
  HTTPUpload& upload = server.upload();
  if (upload.status == UPLOAD_FILE_START) {
    if (SD.exists((char *)upload.filename.c_str())) {
      SD.remove((char *)upload.filename.c_str());
    }
    uploadFile = SD.open(upload.filename.c_str(), FILE_WRITE);
    DBG_OUTPUT_PORT.print("Upload: START, filename: "); DBG_OUTPUT_PORT.println(upload.filename);
  } else if (upload.status == UPLOAD_FILE_WRITE) {
    if (uploadFile) {
      uploadFile.write(upload.buf, upload.currentSize);
    }
    DBG_OUTPUT_PORT.print("Upload: WRITE, Bytes: "); DBG_OUTPUT_PORT.println(upload.currentSize);
  } else if (upload.status == UPLOAD_FILE_END) {
    if (uploadFile) {
      uploadFile.close();
    }
    DBG_OUTPUT_PORT.print("Upload: END, Size: "); DBG_OUTPUT_PORT.println(upload.totalSize);
  }
}

void deleteRecursive(String path) {
  File file = SD.open((char *)path.c_str());
  if (!file.isDirectory()) {
    file.close();
    SD.remove((char *)path.c_str());
    return;
  }

  file.rewindDirectory();
  while (true) {
    File entry = file.openNextFile();
    if (!entry) {
      break;
    }
    String entryPath = path + "/" + entry.name();
    if (entry.isDirectory()) {
      entry.close();
      deleteRecursive(entryPath);
    } else {
      entry.close();
      SD.remove((char *)entryPath.c_str());
    }
    yield();
  }

  SD.rmdir((char *)path.c_str());
  file.close();
}

void handleDelete() {
  if (server.args() == 0) {
    return returnFail("BAD ARGS");
  }
  String path = server.arg(0);
  if (path == "/" || !SD.exists((char *)path.c_str())) {
    returnFail("BAD PATH");
    return;
  }
  deleteRecursive(path);
  returnOK();
}

void handleCreate() {
  if (server.args() == 0) {
    return returnFail("BAD ARGS");
  }
  String path = server.arg(0);
  if (path == "/" || SD.exists((char *)path.c_str())) {
    returnFail("BAD PATH");
    return;
  }

  if (path.indexOf('.') > 0) {
    File file = SD.open((char *)path.c_str(), FILE_WRITE);
    if (file) {
      file.write(0);
      file.close();
    }
  } else {
    SD.mkdir((char *)path.c_str());
  }
  returnOK();
}

void printDirectory() {
  if (!server.hasArg("dir")) {
    return returnFail("BAD ARGS");
  }
  String path = server.arg("dir");
  if (path != "/" && !SD.exists((char *)path.c_str())) {
    return returnFail("BAD PATH");
  }
  File dir = SD.open((char *)path.c_str());
  path = String();
  if (!dir.isDirectory()) {
    dir.close();
    return returnFail("NOT DIR");
  }
  dir.rewindDirectory();
  server.setContentLength(CONTENT_LENGTH_UNKNOWN);
  server.send(200, "text/json", "");
  WiFiClient client = server.client();

  server.sendContent("[");
  for (int cnt = 0; true; ++cnt) {
    File entry = dir.openNextFile();
    if (!entry) {
      break;
    }

    String output;
    if (cnt > 0) {
      output = ',';
    }

    output += "{\"type\":\"";
    output += (entry.isDirectory()) ? "dir" : "file";
    output += "\",\"name\":\"";
    output += entry.name();
    output += "\"";
    output += "}";
    server.sendContent(output);
    entry.close();
  }
  server.sendContent("]");
  dir.close();
}

void handleNotFound() {
  if (hasSD && loadFromSdCard(server.uri())) {
    return;
  }
  String message = "SDCARD Not Detected\n\n";
  message += "URI: ";
  message += server.uri();
  message += "\nMethod: ";
  message += (server.method() == HTTP_GET) ? "GET" : "POST";
  message += "\nArguments: ";
  message += server.args();
  message += "\n";
  for (uint8_t i = 0; i < server.args(); i++) {
    message += " NAME:" + server.argName(i) + "\n VALUE:" + server.arg(i) + "\n";
  }
  server.send(404, "text/plain", message);
  DBG_OUTPUT_PORT.print(message);
}

void initSDcard() {
  master_boot_record_t buffer;
  hspi = new SPIClass(HSPI);
  hspi->begin(HSPI_SCLK, HSPI_MISO, HSPI_MOSI, HSPI_SS); //SCLK, MISO, MOSI, SS

  if (SD.begin(HSPI_SS, *hspi)) {
    DBG_OUTPUT_PORT.println("SD Card initialized.");
    hasSD = true;
  }

  readSD((uint8_t*)&buffer, 0); // read MBR from SD card

  sectors = buffer.partitions[0].sector_count; // we are assuming in this example there is only 1 partition

  Serial.printf("Partition type => %d\n", buffer.partitions[0].partition_type);
  msc.begin();
}

void setup(void) {
  DBG_OUTPUT_PORT.begin(115200);
  DBG_OUTPUT_PORT.setDebugOutput(true);
  DBG_OUTPUT_PORT.print("\n");

  initSDcard();

  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  DBG_OUTPUT_PORT.print("Connecting to ");
  DBG_OUTPUT_PORT.println(ssid);

  // Wait for connection
  uint8_t i = 0;
  while (WiFi.status() != WL_CONNECTED && i++ < 20) {//wait 10 seconds
    delay(500);
  }
  if (i == 21) {
    DBG_OUTPUT_PORT.print("Could not connect to ");
    DBG_OUTPUT_PORT.println(ssid);
    while (1) {
      delay(500);
    }
  }
  DBG_OUTPUT_PORT.print("Connected! IP address: ");
  DBG_OUTPUT_PORT.println(WiFi.localIP());

  if (MDNS.begin(host)) {
    MDNS.addService("http", "tcp", 80);
    DBG_OUTPUT_PORT.println("MDNS responder started");
    DBG_OUTPUT_PORT.print("You can now connect to http://");
    DBG_OUTPUT_PORT.print(host);
    DBG_OUTPUT_PORT.println(".local");
  }


  server.on("/list", HTTP_GET, printDirectory);
  server.on("/edit", HTTP_DELETE, handleDelete);
  server.on("/edit", HTTP_PUT, handleCreate);
  server.on("/edit", HTTP_POST, []() {
    returnOK();
  }, handleFileUpload);
  server.onNotFound(handleNotFound);

  server.begin();
  DBG_OUTPUT_PORT.println("HTTP server started");

}

void loop(void) {
  server.handleClient();
}


/**
 * Functions to read raw sectors from sd card, it requires to update arduino SD library
 */
bool readSD(uint8_t* buffer, uint32_t sector)
{
  return SD.readRAW(buffer, sector);
}

bool writeSD(uint8_t* buffer, uint32_t sector)
{
  return SD.writeRAW(buffer, sector);
}
